/*
 *
 *  Compile with flex
 */



/*** Code inserted at the begin of the file. ***/
%{
#include "Term.h"       // To be included to avoid compiling errors
#include "modal.tab.h"
#include <string.h>
%}


/*** Options ***/
%option noyywrap
%option never-interactive

/*** Regular name definitions ***/
DIGIT                   [0-9]
LETTER                  [a-zA-Z]
WHITE                   [ \t\r\n]+


/*** Rules ***/
%%

{WHITE}                      { } /* Spaces */

"#"[^\n]*\n                  { } /* Comment line */

"true"                       { yylval.bval = true; return VALUE; } 
"false"                      { yylval.bval = false; return VALUE; } 

{DIGIT}+                     { yylval.ival = atoi(yytext); return INT; } /* Integer numbers (truth values) */

{LETTER}[a-zA-Z0-9_]*        { yylval.sval = strdup(yytext); return IDENTIFIER; } /* Identifiers (propositional variables) */

"-"                          { return NOT; } /* Operators (boolean and modal) */
"*"                          { return AND; }
"+"                          { return OR; }
"@"                          { return BOX; }
"^"                          { return DIAMOND; }

","                          { return COMMA; } /* Other tokens */
"("                          { return LROUND; }
")"                          { return RROUND; }

.                            { printf("LEXER: unrecognized token '%s'\n", yytext); exit(-1); } /* Unrecognized text */

<<EOF>>                      { return EOF; } /* End of file */


%%
/*** Code inserted at the and of the file ***/
