/*
 *
 *  Compile with bison.
 */

/*** Code inserted at the begin of the file. ***/
%{   
  #include <stdlib.h>
  #include <list>
  
  #include "Def.h"
  #include "Term.h"
  #include "Formula.h"

  // yylex exists
  extern int yylex();
  extern char yytext[];

  void yyerror(char *msg);
%}


/*** Bison declarations ***/
%union
{
   bool         bval;
   int          ival;
   operator_t  opval;
   char        *sval;
   TermPtr     *term;      
   term_list_t *list;
}

%token COMMA
%token LROUND RROUND 

/* No need of these rules using infix notation 
%left AND OR 
%right NOT BOX DIAMOND 
*/

%token AND OR NOT BOX DIAMOND
%token VALUE
%token IDENTIFIER
%token INT

%type<bval> VALUE
%type<ival> INT
%type<sval> IDENTIFIER 

%type<term> Formula BooleanValue BooleanFormula ModalFormula PropositionalVariable
%type<list> FormulaList
%type<opval> EnnaryBoolOperator UnaryBoolOperator ModalOperator

%start Start

/* Grammar rules */
%%
Start:   /* empty */
       | Formula                    { (Formula::getFormula()).setRoot(*$1); }
;

Formula:   BooleanValue             { $$ = $1; }
         | PropositionalVariable    { $$ = $1; }
         | BooleanFormula           { $$ = $1; }
         | ModalFormula             { $$ = $1; }
;

BooleanValue:   VALUE
                  {
                    $$ = new TermPtr( (Term *) new BooleanValue($1) );
                  }
;

PropositionalVariable:   IDENTIFIER    
                           {
                             $$ = new TermPtr( (Term*) new PropositionalVar($1) );
                           }
;

BooleanFormula:   EnnaryBoolOperator LROUND FormulaList RROUND
                    {
                      $$ = new TermPtr( (Term*) new BooleanOp($3, $1) );
                    }
                | UnaryBoolOperator LROUND Formula RROUND
                    {
                      if ($1 == O_NOT) {($3)->Negate();} $$ = $3; 
                    }
;

EnnaryBoolOperator:   AND     { $$ = O_AND; }
                    | OR      { $$ = O_OR; }
;

UnaryBoolOperator:   NOT      { $$ = O_NOT; }
;

FormulaList:   Formula COMMA Formula
                 {
                   $$ = new term_list_t();
                   $$->push_front(*$3); delete($3);
                   $$->push_front(*$1); delete($1);
                 }
             | Formula COMMA FormulaList
                 {
                   $$ = $3;
                   $$->push_front(*$1); delete($1);
                 }
; 

ModalFormula:   ModalOperator LROUND Formula RROUND
                  {
                    $$ = new TermPtr( (Term*) new ModalOp(*$3, $1) ); 
                    delete($3); 
                  }
;

ModalOperator:   BOX          { $$ = O_BOX; }
               | DIAMOND      { $$ = O_DIAMOND; }
;


/*** Code inserted at the and of the file ***/
%%

void yyerror(char *msg)
{
  printf("PARSER: %s", msg);
  if (yytext[0] != 0)
    printf(" near token '%s'\n", yytext);
  else 
    printf("\n");
  exit(-1);   
}

