/*
 *
 *  Compile with bison.
 */

/*** Code inserted at the begin of the file. ***/
%{   
  #include <stdlib.h>
  #include <list>
  
  #include "Def.h"
  #include "Term.h"
  #include "Formula.h"

  // yylex exists
  extern int yylex();
  extern char yytext[];

  void yyerror(char *msg);
%}


/*** Bison declarations ***/
%union
{
   bool         bval;
   operator_t  opval;
   char        *sval;
   TermPtr     *term;      
}

%token LROUND RROUND 

%left IFF
%left IMP
%left OR
%left AND
%right DIAMOND 
%right BOX 
%right NOT 

%token VALUE
%token IDENTIFIER

%type<bval> VALUE
%type<sval> IDENTIFIER 

%type<term> Formula BooleanValue BooleanFormula ModalFormula PropositionalVariable
%type<opval> BinaryBoolOperator UnaryBoolOperator ModalOperator

%start Start

/* Grammar rules */
%%
Start:   /* empty */
       | Formula                    { (Formula::getFormula()).setRoot(*$1); }
;

Formula:   BooleanValue             { $$ = $1; }
         | PropositionalVariable    { $$ = $1; }
         | BooleanFormula           { $$ = $1; }
         | ModalFormula             { $$ = $1; }
         | LROUND Formula RROUND    { $$ = $2; }
;

BooleanValue:   VALUE
                  {
                    $$ = new TermPtr( (Term*) new BooleanValue($1) );
                  }
;

PropositionalVariable:   IDENTIFIER    
                           {
                             $$ = new TermPtr( (Term*) new PropositionalVar($1) );
                           }
;

BooleanFormula:   Formula BinaryBoolOperator Formula
                    {
/*
                      term_list_t *new_list = new term_list_t();
                      
                      new_list->push_front(*$3); delete($3);
                      new_list->push_front(*$1); delete($1);
                      $$ = new TermPtr( (Term*) new BooleanOp(new_list, $2) );
*/
                      $$ = new TermPtr( (Term*) new BooleanOp(*$1, *$3, $2) );
                      delete($3); delete($1);
                    }
                | Formula IMP Formula
                    {
/*
                      term_list_t *new_list = new term_list_t();
                    
                      ($1)->Negate();
                      new_list->push_front(*$3); delete($3);
                      new_list->push_front(*$1); delete($1);
                      $$ = new TermPtr( (Term*) new BooleanOp(new_list, O_OR) );
*/                  
                      ($1)->Negate();
                      $$ = new TermPtr( (Term*) new BooleanOp(*$1, *$3, O_OR) );
                      delete($3); delete($1);
                  }
                | PropositionalVariable IFF PropositionalVariable
                    {
                      PropositionalVar *Copy1 = new PropositionalVar( *((PropositionalVar*)$1->getPtr()) );
                      PropositionalVar *Copy3 = new PropositionalVar( *((PropositionalVar*)$3->getPtr()) );
                      
                      TermPtr Negated1( (Term*)Copy1, $1->isNegated() ); TermPtr Negated3( (Term*)Copy3, $3->isNegated() );
                      Negated1.Negate(); Negated3.Negate();
                      
                      TermPtr Or1( (Term*) new BooleanOp(Negated1, *$3, O_OR) );
                      TermPtr Or2( (Term*) new BooleanOp(Negated3, *$1, O_OR) );
                      
                      $$ = new TermPtr( (Term*) new BooleanOp(Or1, Or2, O_AND) );
                      delete($3); delete($1);
                    }
                | UnaryBoolOperator Formula 
                    {
                      if ($1 == O_NOT) {($2)->Negate();} $$ = $2;
                    }
;

BinaryBoolOperator:   AND     { $$ = O_AND; }
                    | OR      { $$ = O_OR; }
;

UnaryBoolOperator:   NOT      { $$ = O_NOT; }
;

ModalFormula:   ModalOperator Formula 
                  {
                    $$ = new TermPtr( (Term*) new ModalOp(*$2, $1) );
                    delete($2);
                  }
;

ModalOperator:   BOX          { $$ = O_BOX; }
               | DIAMOND      { $$ = O_DIAMOND; }
;


/*** Code inserted at the and of the file ***/
%%

void yyerror(char *msg)
{
  printf("PARSER: %s", msg);
  if (yytext[0] != 0)
    printf(" near token '%s'\n", yytext);
  else 
    printf("\n");
  exit(-1);   
}

