/*
 *  Modal Logic To Sat Translator
 *
 *  2005-2006
 *  Michele Vescovi
 */
#include "Label.h"



#include <cstdio>



unsigned long Label::unique_id_counter = 0;



Label::Label (Term *creator, Label *parent_label)
{
  // Default initialization (valid for root)
  text = 0;
  label = 1;
  Parent = 0;
  root = true;
  
  // Initializations
  unique_id = ++Label::unique_id_counter;
  modal_creator = creator;

  RootTerms = 0;
  
  // Construct the list of children labels (empty)
  Children = new list<Label *>();
  
  if (parent_label)
    // this label is a child of parent_label (parent_label sets the value of label)
    parent_label->addChildLabel(this);
  // ...else this is the root label
  
  // call to getText on construction, otherwise it will not works if parent will be destroyed
  getText();
}

Label::~Label ()
{
  if (Children)
  {
    label_it_t it;

    // Destroy children and finally the list    
    for (it = Children->begin(); it != Children->end(); ++it)
      delete (*it);
    Children->clear();
    delete Children;
  }
  if (text) delete text;
  if (RootTerms) delete (RootTerms);
}

void Label::destroyThis ()
{
  if (Children)
  {
    // Clear the list of children without destroy children 
    Children->clear();

    // Free the list 
    delete (Children);
    Children = 0;
  }
  if (text) delete (text);
  text = 0;
  if (RootTerms) delete (RootTerms);
  RootTerms = 0;
  delete (this);
}

int Label::addChildLabel (Label *child) 
{ 
  if (child && Children)
  {
    // Adds the new child as the last of all the children
    this->Children->push_back(child);
    // Set the child attributes
    child->Parent = this;
    child->label = this->Children->size();
    child->root = false;
    // Reset the textual description (not needed)
    if (child->text)
    {
      delete (child->text);
      child->text = 0;
    }
    // Returns the label value of the new child    
    return (child->label);
  }
  return 0;
}

const char *Label::getText () 
{
  if (!text)
  {
    // ...this is the first call: the description has to be constructed here
    char tmp[32];

    if (!root) 
    {
      if (! Parent) return 0;
        
      // Use the parent textual descriprion appending _ and this label value
      text = new string(Parent->getText());
      sprintf(tmp, "_%d", label);
      (*text) += tmp;
    }
    else
    {
      // This is root: the description corresponds to the label value
      sprintf(tmp, "%d", label);
      text = new string(tmp);
    }
  }
  // ...else the description has been already constructed
  
  // Return the constant pointer to the textual description
  return text->c_str();
}

