/*
 *  Modal Logic To Sat Translator
 *
 *  2005-2006
 *  Michele Vescovi
 */
#include "Def.h"
#include "Term.h"
#include "Label.h"
#include "Formula.h"



#include <cstdio>



#define DEFAULT_DIAMONDS_ELIMINATION              false
#define DEFAULT_BFS_DFS_TRANSLATION               true
#define DEFAULT_TRANSLATION_T                     T_CNF_DIMACS



#define DEFAULT_MODAL_TRUE_FALSE_ELIMINATION      false
#define DEFAULT_BOOL_ASSOCIATION                  true
#define DEFAULT_MODAL_SIMPLIFYCATION              true
#define DEFAULT_REPEAT_BA_MS                      true



#define DEFAULT_DEBUG                             false



#define F                                         Formula::getFormula()



extern int yyparse();
extern int yydebug;



/* QUI */ 
// 0.  Introdurre gli IFF e XOR ?
// 1.a Usare ClassChunker dove possibile
// 2.  Altre traduzioni (mathsat, limboole, bc)
// 3.  Produzione albero etichette con DOT, o rappresentazione di altro 

// >3. Nodo DAG in cui c'e' figlio normale e figlio not

/* QUI */ 
// 1. solo box no diamonds
// 2. unit propagation
// 3?. BED

int main(int argc, char *argv[])
{
  bool error = false;
  
  bool bfs_dfs, /*NNF*//* nnf, */ diamond_elimation;
  bool modal_true_false_elim, associate_bool, simplify_modal, repeat_ba_ms;
  bool modified_ba, modified_ms, iterate;
  bool debug;
  
  transl_t output_format;
  
  FILE *header_file = stderr;
  FILE *out_file = stdout;
  FILE *var_map_file = 0;
  FILE *debug_file = 0;
  
  bfs_dfs = DEFAULT_BFS_DFS_TRANSLATION;
  
  diamond_elimation = DEFAULT_DIAMONDS_ELIMINATION;
  
  /*NNF*//* nnf = DEFAULT_NNF; */
  modal_true_false_elim = DEFAULT_MODAL_TRUE_FALSE_ELIMINATION;
  associate_bool = DEFAULT_BOOL_ASSOCIATION;
  simplify_modal = DEFAULT_MODAL_SIMPLIFYCATION;
  repeat_ba_ms = DEFAULT_REPEAT_BA_MS;
  
  output_format = DEFAULT_TRANSLATION_T;
  debug = DEFAULT_DEBUG;

  for (int i = 2; i <= argc; ++i)
  {
    if (argv[i-1][0] == '-')
    {
      switch(argv[i-1][1])
      {
        case 'b': diamond_elimation = !DEFAULT_DIAMONDS_ELIMINATION; break;
          
        /*NNF*//* case 'n': nnf = !DEFAULT_NNF; break; */
        case 't': modal_true_false_elim = !DEFAULT_MODAL_TRUE_FALSE_ELIMINATION; break;
        case 'a': associate_bool = !DEFAULT_BOOL_ASSOCIATION; break;      
        case 's': simplify_modal = !DEFAULT_MODAL_SIMPLIFYCATION; break;  
        case 'r': repeat_ba_ms = !DEFAULT_REPEAT_BA_MS; break;    
        
        case 'f': bfs_dfs = !DEFAULT_BFS_DFS_TRANSLATION; break;    
        
//        case 'i': output_format = T_ISCAS; break;      // Output in ISACS format (used by UCSB C-SAT and NoClause)
//        case 'l': output_format = T_LIMBOOLE; break;   // Output in LIMBOOLE format (similar to MATHSAT without variable declarations)
//        case 'b': output_format = T_BC; break;         // Output in the input format of BC
        
//        case 'm': output_format = T_MATHSAT; break;         // Output in MATHSAT format
//        case 'e': output_format = T_NONCNF_EDIMACS; break;  // Output in EDIMACS (NON-CNF) format
        case 'c': output_format = T_CNF_DIMACS; break;        // Output in standard (DIMACS) CNF format
        case 'h': 
        case 'o': 
        case 'v': 
        {
          FILE *fff;
          
          // Permits to specify the file used to output: 
          // ...the translation, the translation header, the mapping between ids and propositional variables.
          if ((argv[i-1][2] == '=') && (argv[i-1][3] != 0)) 
          {
            fff = fopen(argv[i-1]+3, "w");

            switch (argv[i-1][1])
            {
              case 'h': if (fff) header_file = fff; break;
              case 'o': if (fff) out_file = fff; break;
              case 'v': if (fff) var_map_file = fff; break;
            }
            if (!fff)
            {
              fprintf(stderr, "Can not write file: %s\n", argv[i-1]+3); 
              error = true;
            }
          }
          else
          {
            fprintf(stderr, "Invalid parameter: %s\n", argv[i-1]);
            error = true;
          }
          break;
        }
   
        case 'd': // Enable the debug modality
        {        
          debug = !DEFAULT_DEBUG; 
        
          // Permits to specify the file used to output the debug prints...
          if ((argv[i-1][2] == '=') && (argv[i-1][3] != 0))   
          {
            FILE *fff;
   
            fff = fopen(argv[i-1]+3, "w");
            if (!fff)
              { fprintf(stderr, "Can not write file: %s\n", argv[i-1]+3); }
            else
              debug_file = fff;
          }
          else 
            debug_file = stderr;
          
          break;
        }
        default:
        {
          fprintf(stderr, "Invalid parameter: %s\n", argv[i-1]);
          error = true;
          break;
        }
      }
    }
    else 
      if (argv[i-1][0])
      {
        fprintf(stderr, "Invalid parameter: %s\n", argv[i-1]);
        error = true;
      }
  }
  

  if (debug && (debug_file == 0)) debug = !debug;

  
  if (error) exit(-1);
  

  yydebug = 0;
  yyparse();

  
  if (debug) F.printFormulaGraph(debug_file, "Input");
  

  if (diamond_elimation)
  {
    F.treeToBNF();
    if (debug) F.printFormulaGraph(debug_file, "BNF");
  }
  else
  {    
    /*NNF*//* if (nnf) */
    F.treeToNNF();
    if (debug) F.printFormulaGraph(debug_file, "NNF");
  }

  F.treeTrueFalseElimination(modal_true_false_elim);
  if (debug) F.printFormulaGraph(debug_file, "True_False_Elimination");

  // "iterate" is false only at the first iteration
  iterate = false;
  do 
  {
    modified_ms = modified_ba = false;

    // Boolean association doesn't generates other boolean associations
    if (associate_bool)
    {      
      modified_ba = F.treeBoolAssociation();
      if (debug) F.printFormulaGraph(debug_file, "Bool_Association");
    }

    // Modal simplifycation can generate boolean associatio...
    // ...moreover can generate other modal simplyfication if and only if boolean association is disabled
    
    // So it modal simplifycation is necessary at the first iteration
    // ...or if boolean association is disabled...
    // ...or if the formula is changed after the boolean association...
    if ( simplify_modal && ((!iterate) || (!associate_bool) || modified_ba) ) // "modified_ba" is true iff associate_bool is true and the formula is changed
    {      
      modified_ms = F.treeModalSimplifycation();
      if (debug) F.printFormulaGraph(debug_file, "Modal_Simplifycation");
    }

    // Another iteration is needed iff modal simplyfication is enabled
    iterate = simplify_modal && modified_ms;
    iterate = iterate && repeat_ba_ms;
  } 
  while (iterate);

  F.createDAG();
  if (debug) F.printFormulaGraph(debug_file, "DAG");
  
/*  
  if (output_format == T_MATHSAT) fprintf(out_file, "\nVAR\n");
  if ((output_format == T_MATHSAT) || (output_format == T_ISCAS)) Node::ParseTree->printSatVariables(out_file, output_format);
  if (output_format == T_MATHSAT) fprintf(out_file, "\nFORMULA\n");
  if (output_format == T_LIMBOOLE) output_format = T_MATHSAT;
*/
  
  F.translateToSat(out_file, header_file, var_map_file, output_format, bfs_dfs); 

/*  
  if ((output_format == T_CNF_DIMACS) || (output_format == T_NON_CNF)) Node::ParseTree->printSatVariables(cnf_file, output_format);
*/
  
  return 0;
}

