/*
 *
 *  Compile with bison.
 */

/*** Code inserted at the begin of the file. ***/
%{   
  #include <stdlib.h>
  #include <stdio.h>
  #include <string.h>

  #include <string>
  #include <set>

  // yylex exists
  extern int yylex();
  extern char yytext[];

  void yyerror(char *msg);
  
  std::string var_list;
  std::set< std::string > vars;
%}

/*** Bison declarations ***/
%union
{
   bool         bval;
   char        *sval;
}

%token LROUND RROUND 

%left AND OR IMP IFF
%right NOT BOX DIAMOND 

%token VALUE
%token IDENTIFIER

%type<bval> VALUE
%type<sval> IDENTIFIER BinaryBoolOperator UnaryBoolOperator ModalOperator BooleanValue PropositionalVariable BooleanFormula ModalFormula Formula

%start Start

/* Grammar rules */
%%
Start:   /* empty */
       | Formula
         { 
            printf("begin_problem(random_formula).\n");  
            printf("list_of_descriptions.\n");  
            printf(" name({* ??? *}).\n");
            printf(" author({* unknown *}).\n");
            printf(" status(unknown).\n");
            printf(" description({* translated from LWB format (by Michele Vescovi) *}).\n");
            printf("end_of_list.\n");
            printf("list_of_symbols.\n predicates [ (r1,0)");
            printf("%s", var_list.c_str());
            printf(" ].\nend_of_list.\n");
            printf("list_of_special_formulae(conjectures, EML).\n");
            printf(" prop_formula(\n  ");  
            printf("%s", $1);
            printf("\n ).");
            printf("\nend_of_list.\n");
            printf("end_problem.\n");  
            delete $1;
         }
;

Formula:   BooleanValue                     { $$ = $1; }
         | PropositionalVariable            { $$ = $1; }
         | BooleanFormula                   { $$ = $1; }
         | ModalFormula                     { $$ = $1; }
         | LROUND Formula RROUND            { $$ = $2; }
;

BooleanValue:   VALUE
                  {
                    if ($1) 
                      $$ = strdup("true");
                    else 
                      $$ = strdup("false");
                  }
;

PropositionalVariable:   IDENTIFIER     
                           { 
                             std::string Temp($1);
                             
                             if (vars.find(Temp) == vars.end())
                             {
                               vars.insert(Temp);
                               
                               var_list += ", (";
                               var_list += $1;
                               var_list += ",0)";
                             }
                             $$ = $1; 
                           }
;

BooleanFormula:   Formula BinaryBoolOperator Formula
                    {
                      int size = strlen($1) + strlen($2) + strlen($3) + 5;
                      char *temp = new char[size];
                      sprintf(temp, "%s(%s, %s)", $2, $1, $3);
                      delete $1; delete $2; delete $3;
                      $$ = temp;
                    }
                | UnaryBoolOperator Formula 
                    {
                      int size = strlen($1) + strlen($2) + 3;
                      char *temp = new char[size];
                      sprintf(temp, "%s(%s)", $1, $2);
                      delete $1; delete $2;
                      $$ = temp;
                    }
;

BinaryBoolOperator:   AND               { $$ = strdup("and"); }
                    | OR                { $$ = strdup("or"); }
                    | IMP               { $$ = strdup("implies"); }
                    | IFF               { $$ = strdup("equiv"); }
;

UnaryBoolOperator:   NOT                { $$ = strdup("not"); }
;

ModalFormula:   ModalOperator Formula
                  {
                    int size = strlen($1) + strlen($2) + 7;
                    char *temp = new char[size];
                    sprintf(temp, "%s(r1, %s)", $1, $2);
                    delete $1; delete $2;
                    $$ = temp;
                  }
;

ModalOperator:   BOX          { $$ = strdup("box"); }
               | DIAMOND      { $$ = strdup("dia"); }
;


/*** Code inserted at the and of the file ***/
%%

int main() 
{
  yyparse();
}

void yyerror(char *msg)
{
  printf("PARSER: %s", msg);
  if (yytext[0] != 0)
    printf(" near token '%s'\n", yytext);
  else 
    printf("\n");
  exit(-1);   
}

