/*
 *
 *  Compile with bison.
 */

/*** Code inserted at the begin of the file. ***/
%{   
  #include <stdlib.h>
  #include <stdio.h>
  #include <string.h>

  // yylex exists
  extern int yylex();
  extern char yytext[];

  void yyerror(char *msg);
%}

/*** Bison declarations ***/
%union
{
   bool         bval;
   char        *sval;
}

%token LROUND RROUND 

%left AND OR IMP IFF
%right NOT BOX DIAMOND 

%token VALUE
%token IDENTIFIER

%type<bval> VALUE
%type<sval> IDENTIFIER BinaryBoolOperator UnaryBoolOperator ModalOperator BooleanValue PropositionalVariable BooleanFormula ModalFormula Formula

%start Start

/* Grammar rules */
%%
Start:   /* empty */
       | Formula
         { 
           printf("(define-primitive-role R1)\n");
           printf("(define-concept C %s)\n", $1);
           delete $1;
         }
;

Formula:   BooleanValue                     { $$ = $1; }
         | PropositionalVariable            { $$ = $1; }
         | BooleanFormula                   { $$ = $1; }
         | ModalFormula                     { $$ = $1; }
         | LROUND Formula RROUND            { $$ = $2; }
;

BooleanValue:   VALUE
                  {
                    if ($1) 
                      $$ = strdup("*top*");
                    else 
                      $$ = strdup("*bottom*");
                  }
;

PropositionalVariable:   IDENTIFIER     { $$ = $1; }
;

BooleanFormula:   Formula BinaryBoolOperator Formula
                    {
                      int size = strlen($1) + strlen($2) + strlen($3) + 4 + 1;
                      char *temp = new char[size];
                      sprintf(temp, "(%s %s %s)", $2, $1, $3);
                      delete $1; delete $2; delete $3;
                      $$ = temp;
                    }
                | Formula IMP Formula
                    {
                      int size = strlen($1) + strlen($3) + 12 + 1;
                      char *temp = new char[size];
                      sprintf(temp, "(or (not %s) %s)", $1, $3);
                      delete $1; delete $3;
                      $$ = temp;
                    }
                | Formula IFF Formula
                    {
                      int size = (2 * (strlen($1) + strlen($3) + 12)) + 7 + 1;
                      char *temp = new char[size];
                      sprintf(temp, "(and (or (not %s) %s) (or (not %s) %s))", $1, $3, $3, $1);
                      delete $1; delete $3;
                      $$ = temp;
                    }
                | UnaryBoolOperator Formula 
                    {
                      int size = strlen($1) + strlen($2) + 3 + 1;
                      char *temp = new char[size];
                      sprintf(temp, "(%s %s)", $1, $2);
                      delete $1; delete $2;
                      $$ = temp;
                    }
;

BinaryBoolOperator:   AND               { $$ = strdup("and"); }
                    | OR                { $$ = strdup("or"); }
;

UnaryBoolOperator:   NOT                { $$ = strdup("not"); }
;

ModalFormula:   ModalOperator Formula
                  {
                    int size = strlen($1) + strlen($2) + 7;
                    char *temp = new char[size];
                    sprintf(temp, "(%s R1 %s)", $1, $2);
                    delete $1; delete $2;
                    $$ = temp;
                  }
;

ModalOperator:   BOX          { $$ = strdup("all"); }
               | DIAMOND      { $$ = strdup("some"); }
;


/*** Code inserted at the and of the file ***/
%%

int main() 
{
  yyparse();
}

void yyerror(char *msg)
{
  printf("PARSER: %s", msg);
  if (yytext[0] != 0)
    printf(" near token '%s'\n", yytext);
  else 
    printf("\n");
  exit(-1);   
}

