/*
 *
 *  Compile with flex
 */



/*** Code inserted at the begin of the file. ***/
%{
#include "tancs_to_lwb.tab.h"
#include <string.h>
%}


/*** Options ***/
%option noyywrap
%option never-interactive

/*** Regular name definitions ***/
DIGIT                   [0-9]
LETTER                  [a-zA-Z]
WHITE                   [ \t\r\n]+


/*** Rules ***/
%%

{WHITE}                      { }                    /* Spaces */

"%"[^\n]*\n                  { }                    /* Comment line */

"inputformula"               { return INPUTFML; }   /* Used in the formula's header */
"conjecture"                 { return CONJ; }
"hypothesis"                 { return HYP; }

"true"                       { yylval.bval = true; return VALUE; } 
"false"                      { yylval.bval = false; return VALUE; } 

"~"                          { return NOT; }        /* Operators (boolean and modal) */
"&"                          { return AND; }
"|"                          { return OR; }
"=>"                         { return IMP; }
"<="                         { return IIMP; }
"<=>"                        { return IFF; }
"<~>"                        { return IIFF; }
"box"                        { return BOX; }
"pos"                        { return DIAMOND; }

"("                          { return LROUND; }     /* Punctation */
")"                          { return RROUND; }
":"                          { return DDOT; } 
"."                          { return DOT; }
","                          { return COMMA; }

[v]{DIGIT}+                  { yylval.sval = strdup(yytext); return VARIABLE; }     /* Propositional variables */
[r]{DIGIT}+                  { yylval.sval = strdup(yytext); return RELATION; }     /* Modalities */

{LETTER}[a-zA-Z0-9_]*        { yylval.sval = strdup(yytext); return IDENTIFIER; }   /* Identifiers (in the formula's header) */

.                            { printf("LEXER: unrecognized token '%s'\n", yytext); exit(-1); }    /* Unrecognized text */

<<EOF>>                      { return EOF; }        /* End of file */


%%
/*** Code inserted at the and of the file ***/
