/*
 *
 *  Compile with bison.
 */

/*** Code inserted at the begin of the file. ***/
%{   
  #include <stdlib.h>
  #include <stdio.h>
  #include <string.h>

  // yylex exists
  extern int yylex();
  extern char yytext[];

  void yyerror(char *msg);

  char *get_new_modal_string(char *op, char *rel, char *fml);
  char *get_new_bool_unary_string(char *op, char *fml);
  char *get_new_bool_binary_string(char *op, char *fml_1, char *fml_2);

  bool newline;  
%}

/*** Bison declarations ***/
%union
{
   bool         bval;
   char        *sval;
}

%left AND
%left OR
%left IFF IIFF
%left IMP IIMP
%left BinaryBoolOperator

%right BOX
%right DIAMOND
%right NOT

%token LROUND RROUND DDOT COMMA DOT INPUTFML HYP CONJ

%token VALUE
%token VARIABLE IDENTIFIER RELATION

%type<bval> VALUE Conjecture
%type<sval> IDENTIFIER VARIABLE RELATION BooleanValue PropositionalVariable BooleanFormula ModalFormula Formula Hypothesis HypothesisList

%start Start

/* Grammar rules */
%%
Start:   /* empty */
       |  HypothesisList DOT Conjecture DOT
            { 
              if ($3)
                yyerror("The problem contains a \"true value\" conjecture, so it is trivially unsatisfiable.\n");
              else
                printf("%s\n",$1);
                
              delete $1;
            }
;

HypothesisList:   Hypothesis                      
                    { $$ = $1; }
                | HypothesisList DOT Hypothesis   
                    { 
                      char *temp;
                      int size = strlen($1) + strlen($3) + 2 + 2 + 1  +1;
                      if (newline) size += 2;
                      temp = new char[size];

                      if (newline)
                        sprintf(temp, "(%s \n & %s)", $1, $3);
                      else
                        sprintf(temp, "(%s & %s)", $1, $3);
                        
                      delete $1; 
                      delete $3;
                      
                      $$ = temp;
                    }
;

Hypothesis:   INPUTFML LROUND IDENTIFIER COMMA HYP COMMA Formula RROUND
                { delete $3; $$ = $7; }
;

Conjecture:   INPUTFML LROUND IDENTIFIER COMMA CONJ COMMA VALUE RROUND
                { delete $3; $$ = $7; }
;

Formula:   BooleanValue                         { $$ = $1; }
         | PropositionalVariable                { char *temp = $1; temp[0] = 'p'; $$ = temp; }
         | BooleanFormula                       { $$ = $1; }
         | ModalFormula                         { $$ = $1; }
         | LROUND Formula RROUND                { $$ = $2; }
;

BooleanValue:   VALUE
                  {
                    if ($1) 
                      $$ = strdup("true");
                    else 
                      $$ = strdup("false");
                  }
;

PropositionalVariable:   VARIABLE               { $$ = $1; }
;

BooleanFormula:   NOT Formula                   { $$ = get_new_bool_unary_string("~", $2); }
                | Formula OR Formula            { $$ = get_new_bool_binary_string("v", $1, $3); }
                | Formula AND Formula           { $$ = get_new_bool_binary_string("&", $1, $3); }
                | Formula IMP Formula           { $$ = get_new_bool_binary_string("->", $1, $3); }
                | Formula IIMP Formula          { $$ = get_new_bool_binary_string("->", $3, $1); }
                | Formula IFF Formula           { $$ = get_new_bool_binary_string("<->", $1, $3); }
                | Formula IIFF Formula          { yyerror("Operator \"<~>\" not supported.\n"); $$ = 0; }
;

ModalFormula:   BOX RELATION DDOT Formula       { $$ = get_new_modal_string("box", $2, $4); }
              | DIAMOND RELATION DDOT Formula   { $$ = get_new_modal_string("dia", $2, $4); }
;



/*** Code inserted at the and of the file ***/
%%

int main(int argc, char *argv[]) 
{
  newline = false;
  if ((argc > 1) && (strcmp(argv[1],"-n") == 0))
    newline = true;
  yyparse();
}


void yyerror(char *msg)
{
  printf("PARSER: %s", msg);
  if (yytext[0] != 0)
    printf(" near token '%s'\n", yytext);
  else 
    printf("\n");
  exit(-1);   
}


char *get_new_modal_string (char *op, char *rel, char *fml)
{
  int size = strlen(op) + strlen(fml) + 2 + 1   +1;
  char *temp = new char[size];
  
  sprintf(temp, "(%s %s)", op, fml);
  delete rel; 
  delete fml;
  
  return temp; 
}


char *get_new_bool_unary_string (char *op, char *fml)
{
  int size = strlen(op) + strlen(fml) + 2   +1;
  char *temp = new char[size];
  
  sprintf(temp, "(%s%s)", op, fml);

  delete fml;
  
  return temp;
}


char *get_new_bool_binary_string (char *op, char *fml_1, char *fml_2)
{
  int size = strlen(fml_1) + strlen(op) + strlen(fml_2) + 2 + 2   +1;
  char *temp = new char[size];

  sprintf(temp, "(%s %s %s)", fml_1, op, fml_2);
  
  delete fml_1; 
  delete fml_2;
  
  return temp;
}
