#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>



#define NOT             "~"
#define AND             "&"
#define OR              "v"
#define BOX             "box"
#define DIAMOND         "dia"



#define DONT_REPEAT_PROP  0



typedef unsigned int uint;



uint gen_uint (uint range);
uint gen_bit_uint (double p_is_1);
uint gen_floor_or_ceil (double C);

void random_formula (FILE *f, uint L, uint d, uint m, uint N, double p, double C);
void random_clause (FILE *f, uint d, uint m, uint N, double p, double C);

void random_modal_atom (FILE *f, uint d, uint m, uint N, double p, double C);
void random_propositional_atom (FILE *f, uint N, uint *dont_repeat);

void random_negation (FILE *f);



uint gen_uint (uint range)
{ 
  //rand() never reach RAND_MAX 
  return (uint)(((double)rand()/(double)RAND_MAX) * (double)(range)); 
} 

uint gen_bit_uint (double p_is_1)
{
  if ( ((double)rand()/(double)RAND_MAX) < p_is_1 )
    return 1;
  else
    return 0;
}

uint gen_floor_or_ceil (double C)
{
  double ceil_C, floor_C;
  uint inc = 0;
  
  ceil_C = ceil(C);
  floor_C = floor(C);
  
  if ( ((uint)(ceil_C - floor_C)) != 0 )
    inc = gen_bit_uint(C - floor_C);  
  
  return (((uint)floor_C) + inc);
}

void random_formula (FILE *f, uint L, uint d, uint m, uint N, double p, double C)
{
  uint i;
  
  fprintf(f, "(\n   ");  
  for (i = 0; i < L; ++i)
  {
    random_clause(f, d, m, N, p, C);
    if (i < (L-1))
      fprintf(f, "\n %s ", AND);
  }
  fprintf(f, "\n)\n"); 
}

void random_clause (FILE *f, uint d, uint m, uint N, double p, double C)
{
  uint i, clause_len, propositional_atoms, modal_atoms;
  uint *already_used;
  
  already_used = (uint*)calloc(N, sizeof(uint));
  
  clause_len = gen_floor_or_ceil(C);
  if (d != 0)
    propositional_atoms = gen_floor_or_ceil(p * ((double)clause_len));
  else 
    propositional_atoms = clause_len;
  modal_atoms = clause_len - propositional_atoms;
  
  fprintf(f, "( ");

  for (i = 0; i < propositional_atoms; ++i)
  {
    random_negation(f);
    if (DONT_REPEAT_PROP)
      random_propositional_atom (f, N, already_used);
    else
      random_propositional_atom (f, N, 0);
    if (i < (propositional_atoms-1))
      fprintf(f, " %s ", OR);
  }
  
  if ( (modal_atoms > 0) && (propositional_atoms > 0) )
    fprintf(f, " %s ", OR);
  
  for (i = 0; i < modal_atoms; ++i)
  {
    random_negation(f);
// se non volessi ripetizioni qui dovrei costruirmi la striga e confrontarla con l'altra...
// ...dovrei quaindi ordinare anche i literal in qualche modo...    
// ...la stringa dovrebbe essere grande 4-6ch [parentesi+spazi] + d * ((floor_C * 4)[spazi e operatore] + (ceil_C * MAX{una box e un prop}))
    random_modal_atom (f, d, m, N, p, C);
    if (i < (modal_atoms-1))
      fprintf(f, " %s ", OR);
  }
  
  fprintf(f, " )");  
  
  free(already_used);
}

void random_negation (FILE *f)
{
  if (gen_bit_uint(0.5) == 1)
    fprintf(f, "%s", NOT);
}

void random_modal_atom (FILE *f, uint d, uint m, uint N, double p, double C)
{
  uint r = 1 + gen_uint(m);
  if (r == 1)
    fprintf(f, "%s", BOX);
  else
    fprintf(f, "r%d: %s", r, BOX);
  random_clause (f, d-1, m, N, p, C);
}

void random_propositional_atom (FILE *f, uint N, uint *dont_repeat)
{
  uint index = gen_uint(N);
    
  if (dont_repeat)
  {
    while (dont_repeat[index])
      index = gen_uint(N);
    dont_repeat[index] = 1;
  }
  fprintf(f, "p%d", index);
}

int main (int argc, char *argv[])
{
  uint seed = 0, d = 1, m = 1, N = 1, L = 10;
  double C = 3.0, p = 0.5;
  int i;

  for (i = 1; i < argc; ++i)
  {
    switch (argv[i][0])
    {
      case 'd':
        d = atoi(argv[i]+1);
        break;

      case 'm':
        m = atoi(argv[i]+1);
        break;

      case 'L':
        L = atoi(argv[i]+1);
        break;

      case 'N':
        N = atoi(argv[i]+1);
        break;

      case 's':
        seed = atoi(argv[i]+1);
        break;

      case 'C':
        C = atof(argv[i]+1);
        break;

      case 'p':
        p = atof(argv[i]+1);
        break;
     
      default:
        fprintf(stderr, "%s: Invalid parmeter '%s'\n", argv[0], argv[i]);
        exit(-1);
    }
  }

  if (seed == 0)
    srand(time(0));
  else 
    srand(seed);
  
  random_formula(stdout, L, d, m, N, p, C);
  
  return 0;
}
