#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include <string>
#include <list>

using namespace std;



#define NOT       "~"
#define AND       " & "
#define OR        " v "
#define BOX       "box"
#define DIAMOND   "dia"




#define DEBUG     false



typedef unsigned int uint;
bool debug = DEBUG;



uint gen_uint (uint range);
uint gen_bit_uint (double p_is_1);
uint gen_floor_or_ceil (double C);

string *random_formula (uint L, uint d, uint m, uint N, double p, double C);
string *random_clause (uint d, uint m, uint N, double p, double C);

string *random_modal_atom (uint d, uint m, uint N, double p, double C);
string *random_propositional_atom (uint N);



uint gen_uint (uint range)
{ 
  //rand() never reach RAND_MAX 
  return (uint)(((double)rand()/(double)RAND_MAX) * (double)(range)); 
} 

uint gen_bit_uint (double p_is_1)
{
  if ( ((double)rand()/(double)RAND_MAX) < p_is_1 )
    return 1;
  else
    return 0;
}

uint gen_floor_or_ceil (double C)
{
  double ceil_C, floor_C;
  uint inc = 0;
  
  ceil_C = ceil(C);
  floor_C = floor(C);
  
  if ( ((uint)(ceil_C - floor_C)) != 0 )
    inc = gen_bit_uint(C - floor_C);  
  
  return (((uint)floor_C) + inc);
}

bool insert (string *new_string, list<string *> *strings)
{
  list<string *>::iterator s = strings->begin();
  int res;
  
  while (s != strings->end())
  {
    res = new_string->compare(**s);
    if (res)
      if (res > 0)
        ++s;
      else
      {
        strings->insert(s, new_string);
        return true;
      }
    else
      return false;
  }
  strings->push_back(new_string);
  return true;
}

string *random_formula (uint L, uint d, uint m, uint N, double p, double C)
{
  list<string *>::iterator c;
  list<string *> clauses;
  string *clause, *formula;
  uint i;
  
  i = 0;
  while (i < L)
  {
    clause = random_clause(d, m, N, p, C);
    if (insert(clause, &clauses))
      ++i;
    else
    { 
      if (debug)
        fprintf(stderr, "[%d] clausola ripetuta nella formula: %s\n", i, clause->c_str());
      delete clause;
    }
  }

  formula = new string();
  
  *formula += "(\n   ";  
  for (c = clauses.begin(); c != clauses.end(); ++c)
  {
    if (c != clauses.begin())
    {
      *formula += "\n";
      *formula += AND;
    }
    *formula += **c;
    delete *c;
  }
  *formula += "\n)"; 
  
  return formula;
}

string *random_clause (uint d, uint m, uint N, double p, double C)
{
  uint i, clause_len, prop_atoms_num, modal_atoms_num;
  list<string *> prop_atoms, modal_atoms;
  list<string *>::iterator a;
  string *atom, *clause;
  
  clause_len = gen_floor_or_ceil(C);
  if (d != 0)
    prop_atoms_num = gen_floor_or_ceil(p * ((double)clause_len));
  else 
    prop_atoms_num = clause_len;
  modal_atoms_num = clause_len - prop_atoms_num;
  
  i = 0; 
  while (i < prop_atoms_num)
  {
    atom = random_propositional_atom (N);
    if (insert(atom, &prop_atoms))
      ++i;
    else
    { 
      if (debug)
        fprintf(stderr, "[%d] atomo prop. ripetuto nella clausola: %s\n", prop_atoms.size(), atom->c_str());
      delete atom;
    }
  }
  
  i = 0; 
  while (i < modal_atoms_num)
  {
    atom = random_modal_atom (d, m, N, p, C);
    if (insert(atom, &modal_atoms))
      ++i;
    else
    { 
      if (debug)
        fprintf(stderr, "[%d,%d] atomo modale ripetuto nella clausola: %s\n", prop_atoms.size(), modal_atoms.size(), atom->c_str());
      delete atom;
    }
  }

  prop_atoms.splice(prop_atoms.end(), modal_atoms);
  
  clause = new string();
  *clause += "( ";
  for (a = prop_atoms.begin(); a != prop_atoms.end(); ++a)
  {
    if (a != prop_atoms.begin())
      *clause += OR;
    if (gen_bit_uint(0.5) == 1)
      *clause += NOT;
    *clause += **a;
    delete *a;
  }
  *clause += " )"; 

  return clause;  
}

string *random_modal_atom (uint d, uint m, uint N, double p, double C)
{
  char buffer[1024];
  string *atom, *clause;
  
  uint r = 1 + gen_uint(m);
  if (r == 1)
    sprintf(buffer, "%s", BOX);
  else
    sprintf(buffer, "r%d:%s", r, BOX);
  
  atom = new string(buffer);
  clause = random_clause (d-1, m, N, p, C);
  *atom += *clause;
  delete clause;
  
  return atom;
}

string *random_propositional_atom (uint N)
{
  char buffer[1024];
  
  uint index = gen_uint(N);
  sprintf(buffer, "p%d", index);

  return (new string(buffer));
}

int main (int argc, char *argv[])
{
  uint seed = 0, d = 1, m = 1, N = 1, L = 10;
  double C = 3.0, p = 0.5;
  string *formula;
  int i;

  for (i = 1; i < argc; ++i)
  {
    switch (argv[i][0])
    {
      case 'd':
        d = atoi(argv[i]+1);
        break;

      case 'm':
        m = atoi(argv[i]+1);
        break;

      case 'L':
        L = atoi(argv[i]+1);
        break;

      case 'N':
        N = atoi(argv[i]+1);
        break;

      case 's':
        seed = atoi(argv[i]+1);
        break;

      case 'C':
        C = atof(argv[i]+1);
        break;

      case 'p':
        p = atof(argv[i]+1);
        break;

      case 'e':
        if (argv[i][1] == 0)
        {
          debug = !DEBUG;
          break;
        }
        
      default:
        fprintf(stderr, "%s: Invalid parmeter '%s'\n", argv[0], argv[i]);
        exit(-1);
    }
  }

  if (seed == 0)
    srand(time(0));
  else 
    srand(seed);
  
  formula = random_formula(L, d, m, N, p, C);
  fprintf(stdout, "%s\n", formula->c_str());
  delete formula;
  
  return 0;
}
