#include <cstdlib>
#include <cstring>
#include <cstdio>
#include <cmath>
#include <ctime>

#include <string>
#include <list>

using namespace std;



#define DEBUG               false
#define NEWLINE             false
#define SEED_FILE_NAME      "./seed"



#define DEFAULT_L           10
#define DEFAULT_d           1
#define DEFAULT_m           1
#define DEFAULT_N           5
#define DEFAULT_C           3.0
#define DEFAULT_p           0.5


#define NOT                 "not"
#define AND                 "and"
#define OR                  "or"

#define NOT_S                "~"
#define AND_S               " & "
#define OR_S                " v "

#define BOX                 "box"
#define ALL                 "all"

#define DIAMOND             "dia"
#define SOME                "some"



typedef unsigned int uint;

char *valid_formats[3] = { "lwb", "racer", "mspass" };
uint valid_formats_num = 3;
enum format_t 
{
  LWB_F = 0,
  RACER_F,
  MSPASS_F
};

format_t format = (format_t)0;



bool debug = DEBUG;
bool newline = NEWLINE;



uint gen_uint (uint range);
uint gen_bit_uint (double p_is_1);
uint gen_floor_or_ceil (double C);

bool insert (string *new_string, list<string *> *strings);

string *random_formula (uint L, uint d, uint m, uint N, double p, double C);
string *random_clause (uint d, uint m, uint N, double p, double C);

string *random_modal_atom (uint d, uint m, uint N, double p, double C);
string *random_propositional_atom (uint N);

string *create_lwb_formula_string (list<string *> *clauses);
string *create_lwb_clause_string (list<string *> *atoms);
string *create_lwb_modality_string  (uint r);

string *create_racer_formula_string (list<string *> *clauses, uint m);
string *create_racer_clause_string (list<string *> *atoms);
string *create_racer_modality_string  (uint r);

string *create_mspass_formula_string (list<string *> *clauses,  uint m, uint N);
string *create_mspass_clause_string (list<string *> *atoms);
string *create_mspass_modality_string  (uint r);

uint gen_uint (uint range)
{ 
  //rand() never reach RAND_MAX 
  return (uint)(((double)rand()/(double)RAND_MAX) * (double)(range)); 
} 

uint gen_bit_uint (double p_is_1)
{
  if ( ((double)rand()/(double)RAND_MAX) < p_is_1 )
    return 1;
  else
    return 0;
}

uint gen_floor_or_ceil (double C)
{
  double ceil_C, floor_C;
  uint inc = 0;
  
  ceil_C = ceil(C);
  floor_C = floor(C);
  
  if ( ((uint)(ceil_C - floor_C)) != 0 )
    inc = gen_bit_uint(C - floor_C);  
  
  return (((uint)floor_C) + inc);
}

bool insert (string *new_string, list<string *> *strings)
{
  list<string *>::iterator s = strings->begin();
  int res;
  
  while (s != strings->end())
  {
    res = new_string->compare(**s);
    if (res)
      if (res > 0)
        ++s;
      else
      {
        strings->insert(s, new_string);
        return true;
      }
    else
      return false;
  }
  strings->push_back(new_string);
  return true;
}

string *random_formula (uint L, uint d, uint m, uint N, double p, double C)
{
  list<string *> clauses;
  string *clause, *formula;
  uint i;
  
  i = 0;
  while (i < L)
  {
    clause = random_clause(d, m, N, p, C);
    if (insert(clause, &clauses))
      ++i;
    else
    { 
      if (debug)
        fprintf(stderr, "[%d] clausola ripetuta nella formula: %s\n", i, clause->c_str());
      delete clause;
    }
  }

  switch (format)
  {
    case LWB_F: formula = create_lwb_formula_string(&clauses); break;
    case RACER_F: formula = create_racer_formula_string(&clauses, m); break;
    case MSPASS_F: formula = create_mspass_formula_string(&clauses, m, N); break;
  }
  return formula;
}

string *random_clause (uint d, uint m, uint N, double p, double C)
{
  uint i, clause_len, prop_atoms_num, modal_atoms_num;
  list<string *> prop_atoms, modal_atoms;
  string *atom, *clause;
  
  clause_len = gen_floor_or_ceil(C);
  if (d != 0)
    prop_atoms_num = gen_floor_or_ceil(p * ((double)clause_len));
  else 
    prop_atoms_num = clause_len;
  modal_atoms_num = clause_len - prop_atoms_num;
  
  i = 0; 
  while (i < prop_atoms_num)
  {
    atom = random_propositional_atom (N);
    if (insert(atom, &prop_atoms))
      ++i;
    else
    { 
      if (debug)
        fprintf(stderr, "[%d] atomo prop. ripetuto nella clausola: %s\n", prop_atoms.size(), atom->c_str());
      delete atom;
    }
  }
  
  i = 0; 
  while (i < modal_atoms_num)
  {
    atom = random_modal_atom (d, m, N, p, C);
    if (insert(atom, &modal_atoms))
      ++i;
    else
    { 
      if (debug)
        fprintf(stderr, "[%d,%d] atomo modale ripetuto nella clausola: %s\n", prop_atoms.size(), modal_atoms.size(), atom->c_str());
      delete atom;
    }
  }

  prop_atoms.splice(prop_atoms.end(), modal_atoms);
  switch (format)
  {
    case LWB_F: clause = create_lwb_clause_string(&prop_atoms); break;
    case RACER_F: clause = create_racer_clause_string(&prop_atoms); break;
    case MSPASS_F: clause = create_mspass_clause_string(&prop_atoms); break;
  }
  return clause;  
}

string *random_modal_atom (uint d, uint m, uint N, double p, double C)
{
  string *atom, *clause;
  
  uint r = 1 + gen_uint(m);
  
  switch (format)
  {
    case LWB_F: atom = create_lwb_modality_string(r); break;
    case RACER_F: atom = create_racer_modality_string(r); break;
    case MSPASS_F: atom = create_mspass_modality_string(r); break;
  }
  
  clause = random_clause (d-1, m, N, p, C);
  *atom += *clause;
  delete clause;
  *atom += ")";
  
  return atom;
}

string *random_propositional_atom (uint N)
{
  char buffer[1024];
  
  uint index = gen_uint(N);
  sprintf(buffer, "p%d", index);

  return (new string(buffer));
}



string *create_lwb_clause_string (list<string*> *atoms)
{
  list<string *>::iterator a;
  string *clause = new string();
  
  for (uint j = 1; j < atoms->size(); ++j)
    *clause += "(";
  for (a = atoms->begin(); a != atoms->end(); ++a)
  {
    if (a != atoms->begin())
      *clause += OR_S;
    if (gen_bit_uint(0.5) == 1)
    {
      *clause += "(";
      *clause += NOT_S;
      *clause += **a + ")";
    }
    else
      *clause += **a;
    delete *a;
    if (a != atoms->begin())
      *clause += ")";
  }
  return clause;  
}

string *create_lwb_formula_string (list<string *> *clauses)
{
  list<string *>::iterator c;
  string *formula = new string();
  
  for (uint j = 1; j < clauses->size(); ++j)
    *formula += "( ";
  if (newline) *formula += "\n   ";
  for (c = clauses->begin(); c != clauses->end(); ++c)
  {
    if (c != clauses->begin())
    {
      if (newline) *formula += "\n";
      *formula += AND_S;
      *formula += **c;
      *formula += " )";
    }
    else
      *formula += **c;
    delete *c;
  }
  return formula;
}

string *create_lwb_modality_string (uint r)
{
  char buffer[1024];
  string *modality;

  if (r == 1)
    sprintf(buffer, "(%s", BOX);
  else
    sprintf(buffer, "(r%d: %s", r, BOX);
  
  modality = new string(buffer);
  return modality;
}



string *create_racer_clause_string (list<string*> *atoms)
{
  list<string *>::iterator a;
  string *clause = new string();
  
  *clause += "(";
  *clause += OR;
  for (a = atoms->begin(); a != atoms->end(); ++a)
  {
    if (gen_bit_uint(0.5) == 1)
    {
      *clause += " (";
      *clause += NOT;
      *clause += " ";
      *clause += **a + ")";
    }
    else
      *clause += " " + **a;
    delete *a;
  }
  *clause += ")";
  return clause;  
}

string *create_racer_formula_string (list<string *> *clauses, uint m)
{
  list<string *>::iterator c;
  string *formula = new string();
  char buffer[1024];

  for (uint j = 1; j <= m; ++j)
  {
    sprintf(buffer, "(define-primitive-role R%d) ", j);
    *formula += buffer;
    if (newline) *formula += "\n";
  }
  *formula += "(define-concept C ";  
  if (newline) *formula += "\n";
  *formula += " (";
  *formula += AND;
  for (c = clauses->begin(); c != clauses->end(); ++c)
  {
    if (newline) *formula += "\n";
    *formula += " " + **c;
    delete *c;
  }
  if (newline) *formula += "\n";
  *formula += " )";  
  if (newline) *formula += "\n";  
  *formula += ")";  
  return formula;
}

string *create_racer_modality_string (uint r)
{
  char buffer[1024];
  string *modality;

  sprintf(buffer, "(%s R%d ", ALL, r);
  modality = new string(buffer);
  return modality;
}



string *create_mspass_clause_string (list<string*> *atoms)
{
  list<string *>::iterator a;
  string *clause = new string();

  *clause += " ";
  *clause += OR;
  *clause += "(";
  for (a = atoms->begin(); a != atoms->end(); ++a)
  {
    if (a != atoms->begin())
      *clause += ", ";
    if (gen_bit_uint(0.5) == 1)
    {
      *clause += NOT;
      *clause += "(";
      *clause += **a + ")";
    }
    else
      *clause += **a;
    delete *a;
  }
  *clause += ")";
  return clause;  
}

string *create_mspass_formula_string (list<string *> *clauses, uint m, uint N)
{
  list<string *>::iterator c;
  string *formula = new string();
  char buffer[1024];

  *formula += "begin_problem(random_formula).\n";  
  if (newline) *formula += "\n";
  *formula += "list_of_descriptions.\n";  
  *formula += " name({* ??? *})."; if (newline) *formula += "\n";
  *formula += " author({* kCNFgenerator by Michele Vescovi *})."; if (newline) *formula += "\n";
  *formula += " status(unknown)."; if (newline) *formula += "\n";
  *formula += " description({* randomly generated cnf modal formula *}).\n";
  *formula += "end_of_list.\n";
  if (newline) *formula += "\n";
  *formula += "list_of_symbols.\n predicates [ ";  
  for (uint j = 1; j <= m; ++j)
  {
    if (j != 1)*formula += ", ";
    sprintf(buffer, "(r%d,0)", j);
    *formula += buffer;
  }
  for (uint j = 0; j < N; ++j)
  {
    if ((m > 0) || (j != 0)) *formula += ", ";
    sprintf(buffer, "(p%d,0)", j);
    *formula += buffer;
  }
  *formula += " ].\nend_of_list.\n";
  if (newline) *formula += "\n";
  *formula += "list_of_special_formulae(conjectures, EML).\n";  
  *formula += " prop_formula( ";  
  if (newline) *formula += "\n  ";
  *formula += NOT;
  *formula += "( ";
  *formula += AND;
  *formula += "( ";
  for (c = clauses->begin(); c != clauses->end(); ++c)
  {
    if (newline) *formula += "\n  ";
    if (c != clauses->begin())
      *formula += ", ";
    else 
      *formula += "  ";
    *formula += **c;
    delete *c;
  }
  if (newline) *formula += "\n  ";
  *formula += ") ) ";  
  if (newline) *formula += "\n ";  
  *formula += "). ";  
  *formula += "\nend_of_list.\n";
  if (newline) *formula += "\n";
  *formula += "end_problem.";  
  return formula;
}

string *create_mspass_modality_string (uint r)
{
  char buffer[1024];
  string *modality;

  sprintf(buffer, "%s(r%d,", BOX, r);
  modality = new string(buffer);
  return modality;
}



int main (int argc, char *argv[])
{
  bool use_seed = false;
  uint seed = 0, 
       test = 0, 
       d = DEFAULT_d, 
       m = DEFAULT_m, 
       N = DEFAULT_N, 
       L = DEFAULT_L;
  double C = DEFAULT_C, 
         p = DEFAULT_p;

  char *out_name = 0;
  char *format_name = valid_formats[format];
  char file_name_buffer[1024];
  
  FILE *seed_file, *out_file;
  string *formula;
  int i;

  for (i = 1; i < argc; ++i)
  {
    switch (argv[i][0])
    {
      case 's':
        use_seed= true;
        seed = atoi(argv[i]+1);
        break;

      case 'd':
        d = atoi(argv[i]+1);
        break;

      case 'm':
        m = atoi(argv[i]+1);
        break;

      case 'L':
        L = atoi(argv[i]+1);
        break;

      case 'N':
        N = atoi(argv[i]+1);
        break;

      case 't':
        test = atoi(argv[i]+1);
        break;
      
      case 'C':
        C = atof(argv[i]+1);
        break;

      case 'p':
        p = atof(argv[i]+1);
        break;

      case 'o': case 'f':
        if (argv[i][1] != 0)
        { 
          if (argv[i][0] == 'o') 
          { 
            out_name = argv[i]+1; 
            break; 
          }
          if (argv[i][0] == 'f') 
          {
            uint f;
            
            format_name = argv[i]+1;
            for (f = 0; f < valid_formats_num; ++f)
              if (strcmp(format_name, valid_formats[f]) == 0)
              { 
                format = (format_t)f;
                break;
              }
            if (f != valid_formats_num)
              break;
            else 
              fprintf(stderr, "%s: Option 'f', '%s' is not a supported format\n", argv[0], argv[i]+1);
          }
        } 
        else
          fprintf(stderr, "%s: Invalid option '%s'\n", argv[0], argv[i]);
        exit(-1);
        
      case 'e': case 'n':
        if (argv[i][1] == 0)
        {
          if (argv[i][0] == 'e') debug = !DEBUG;
          if (argv[i][0] == 'n') newline = !NEWLINE;
          break;
        }
        
      default:
        fprintf(stderr, "%s: Invalid option '%s'\n", argv[0], argv[i]);
        exit(-1);
    }
  }

  if (use_seed)
    srand(seed);    
  else 
  {
    srand(seed = (uint)time(0));
    seed_file = fopen(SEED_FILE_NAME, "w");
    fprintf(seed_file, "%d\n", seed);
    fclose(seed_file);
  }

  if (test < 1) 
    test = 1;

  if ((test > 1) && (!out_name))
    newline = false;
  
  if (!out_name) 
    out_file = stdout;
   
  i = 1;
  while (i <= test)
  {
    if (out_name)
    {
      sprintf(file_name_buffer, "%s-%d.%s", out_name, i, format_name);
      out_file = fopen(file_name_buffer, "w");
    }
    formula = random_formula(L, d, m, N, p, C);
    fprintf(out_file, "%s\n", formula->c_str());
    if (out_name) fclose(out_file);
    ++i;
  }
  delete formula;
  
  return 0;
}
